from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from pyoneer.math import logical_ops


class LogicalOpsTest(tf.test.TestCase):
    def test_isclose(self):
        x = tf.constant([0.9, 1.0, 1.1, 1.2])
        actual_x = logical_ops.isclose(x, 1.0, rtol=1e-5, atol=0.1)
        expected_x = tf.constant([True, True, True, False])
        self.assertAllEqual(actual_x, expected_x)


if __name__ == "__main__":
    tf.test.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import gym
import numpy as np
import tensorflow as tf

from pyoneer.rl.wrappers.batch_impl import Batch


class BatchTest(tf.test.TestCase):
    def test_batch_blocking(self):
        batch_size = 8

        env = Batch(
            constructor=lambda: gym.make("Pendulum-v0"),
            batch_size=batch_size,
            blocking=True,
        )
        env.seed(0)

        state = env.reset()
        for _ in range(200):
            action = np.stack(
                [env.action_space.sample() for _ in range(batch_size)], axis=0
            )
            next_state, reward, done, info = env.step(action)

            self.assertTupleEqual(state.shape, (batch_size, 3))
            self.assertTupleEqual(action.shape, (batch_size, 1))
            self.assertTupleEqual(next_state.shape, (batch_size, 3))
            self.assertTupleEqual(reward.shape, (batch_size,))
            self.assertTupleEqual(done.shape, (batch_size,))
            self.assertAllEqual(len(info), batch_size)

    def test_batch_non_blocking(self):
        batch_size = 8

        env = Batch(
            constructor=lambda: gym.make("Pendulum-v0"),
            batch_size=batch_size,
            blocking=False,
        )
        env.seed(0)

        state = env.reset()
        for _ in range(200):
            action = np.stack(
                [env.action_space.sample() for _ in range(batch_size)], axis=0
            )
            next_state, reward, done, info = env.step(action)

            self.assertTupleEqual(state.shape, (batch_size, 3))
            self.assertTupleEqual(action.shape, (batch_size, 1))
            self.assertTupleEqual(next_state.shape, (batch_size, 3))
            self.assertTupleEqual(reward.shape, (batch_size,))
            self.assertTupleEqual(done.shape, (batch_size,))
            self.assertAllEqual(len(info), batch_size)


if __name__ == "__main__":
    tf.test.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf


def isclose(a, b, rtol=1e-5, atol=1e-8):
    """
    Returns a boolean tensor where two arrays are element-wise equal within a
    tolerance.

    The relative difference (rtol * abs(b)) and the absolute difference atol
    are added together to compare against the absolute difference between a
    and b.
    """
    a = tf.convert_to_tensor(a)
    b = tf.convert_to_tensor(b)
    rtol = tf.convert_to_tensor(rtol)
    atol = tf.convert_to_tensor(atol)
    return tf.abs(a - b) <= (atol + rtol * tf.abs(b))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from pyoneer.rl.losses.policy_gradient_ops import (
    PolicyGradient,
    ClippedPolicyGradient,
    PolicyEntropy,
)

__all__ = ["PolicyGradient", "ClippedPolicyGradient", "PolicyEntropy"]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import tensorflow as tf

RADIANS_TO_DEGREES = 180 / math.pi
DEGREES_TO_RADIANS = math.pi / 180


def to_radians(x):
    """
    Convert the inputs from degrees to radians.
    """
    return x * DEGREES_TO_RADIANS


def to_degrees(x):
    """
    Convert the inputs from radians to degrees.
    """
    return x * RADIANS_TO_DEGREES


def to_cartesian(phi, rho=1):
    """
    Convert the inputs from polar to cartesian coordinates.
    """
    x = rho * tf.cos(phi)
    y = rho * tf.sin(phi)
    return x, y


def to_polar(x, y):
    """
    Convert the inputs from cartesian to polar coordinates.
    """
    rho = tf.sqrt(tf.square(x) + tf.square(y))
    phi = tf.atan2(y, x)
    return rho, phi



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow.python.platform import test

from pyoneer.manip import array_ops


class ArrayOpsTest(tf.test.TestCase):
    def test_flatten(self):
        inputs = tf.constant([[0.0, 1.0, 2.0]])
        outputs = array_ops.flatten(inputs)
        expected = tf.constant([0.0, 1.0, 2.0])
        self.assertAllEqual(outputs, expected)

    def test_pad_or_truncate(self):
        inputs = [["apple", "is", "red"], ["i", "dont", "like", "apples"]]

        # truncate
        outputs = tf.stack(
            [
                array_ops.pad_or_truncate(input, sizes=[3], constant_values="<PAD>")
                for input in inputs
            ],
            axis=0,
        )
        expected = tf.constant([["apple", "is", "red"], ["i", "dont", "like"]])
        self.assertAllEqual(outputs, expected)

        # pad
        outputs = tf.stack(
            [
                array_ops.pad_or_truncate(input, sizes=[4], constant_values="<PAD>")
                for input in inputs
            ],
            axis=0,
        )
        expected = tf.constant(
            [["apple", "is", "red", "<PAD>"], ["i", "dont", "like", "apples"]]
        )
        self.assertAllEqual(outputs, expected)

    def test_shift(self):
        # shift forwards
        inputs = tf.constant([[0.0, 1.0, 2.0]])
        outputs = array_ops.shift(inputs, shift=1, axis=1)
        expected = tf.constant([[0.0, 0.0, 1.0]])
        self.assertAllEqual(outputs, expected)

        # shift backwards
        inputs = tf.constant([[0.0, 1.0, 2.0]])
        outputs = array_ops.shift(inputs, shift=-1, axis=1)
        expected = tf.constant([[1.0, 2.0, 0.0]])
        self.assertAllEqual(outputs, expected)

        # shift forwards fill
        inputs = tf.constant([[0.0, 1.0, 2.0]])
        outputs = array_ops.shift(inputs, shift=1, axis=1, padding_values=-1.0)
        expected = tf.constant([[-1.0, 0.0, 1.0]])
        self.assertAllEqual(outputs, expected)

        # shift backwards fill
        inputs = tf.constant([[0.0, 1.0, 2.0]])
        outputs = array_ops.shift(inputs, shift=-1, axis=1, padding_values=3.0)
        expected = tf.constant([[1.0, 2.0, 3.0]])
        self.assertAllEqual(outputs, expected)

        # shift forwards more
        inputs = tf.constant([[0.0, 1.0, 2.0]])
        outputs = array_ops.shift(inputs, shift=2, axis=1)
        expected = tf.constant([[0.0, 0.0, 0.0]])
        self.assertAllEqual(outputs, expected)

        # shift backwards more
        inputs = tf.constant([[0.0, 1.0, 2.0]])
        outputs = array_ops.shift(inputs, shift=-2, axis=1)
        expected = tf.constant([[2.0, 0.0, 0.0]])
        self.assertAllEqual(outputs, expected)

        # shift using tensor constants
        values = tf.constant([[0.0, 0.0, 1.0]])
        bootstrap_values = tf.constant([1.0])
        expected_values_next = tf.concat(
            [values[:, 1:], bootstrap_values[:, None]], axis=1
        )
        actual_values_next = array_ops.shift(
            values, shift=-1, axis=1, padding_values=bootstrap_values[:, None]
        )
        self.assertAllEqual(actual_values_next, expected_values_next)


if __name__ == "__main__":
    test.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import gym
import numpy as np


class ObservationNormalization(gym.Wrapper):
    """
    Wraps the environment to normalize observations.
    """

    def __init__(self, env, mean=None, std=None):
        super(ObservationNormalization, self).__init__(env)

        if mean is None:
            self.mean = (self.observation_space.high + self.observation_space.low) / 2
        else:
            self.mean = mean

        if std is None:
            self.std = (self.observation_space.high - self.observation_space.low) / 2
        else:
            self.std = std

        self.observation_space = self._create_observation_space()

    def _create_observation_space(self):
        low = self.mean - self.std
        high = self.mean + self.std
        return gym.spaces.Box(low, high, dtype=self.env.observation_space.dtype)

    def normalize_observation(self, observ):
        observ = (observ - self.mean) / self.std
        return observ

    def reset(self):
        return self.normalize_observation(self.env.reset())

    def step(self, action):
        observation, reward, done, info = self.env.step(action)
        observation = self.normalize_observation(observation)
        return observation, reward, done, info


class ObservationCoordinates(gym.Wrapper):
    """
    Wraps the environment to append coordinate features.

    Expects the observation space to have shape [height, width, channel]`.
    """

    def __init__(self, env):
        super(ObservationCoordinates, self).__init__(env)
        self.observation_space = self._create_observation_space()

    def _create_observation_space(self):
        observation_space = self.env.observation_space

        coord_low = np.zeros_like(observation_space.low[..., :1])
        coord_high = np.ones_like(observation_space.high[..., :1])

        coords_low = np.tile(coord_low, [1, 1, 3])
        coords_high = np.tile(coord_high, [1, 1, 3])

        low = np.concatenate([observation_space.low, coords_low], axis=-1)
        high = np.concatenate([observation_space.high, coords_high], axis=-1)

        space = gym.spaces.Box(low, high, dtype=observation_space.dtype)
        return space

    def generate_coords(self):
        observation_space = self.observation_space
        x_dim, y_dim = observation_space.shape[:2]

        xx_range = np.linspace(0.0, 1.0, num=x_dim, dtype=observation_space.dtype)
        yy_range = np.linspace(0.0, 1.0, num=y_dim, dtype=observation_space.dtype)

        xx_ones = np.ones(x_dim, dtype=observation_space.dtype)
        yy_ones = np.ones(y_dim, dtype=observation_space.dtype)

        xx_channel = np.matmul(xx_ones[..., None], xx_range[None, ...])
        yy_channel = np.matmul(yy_range[..., None], yy_ones[None, ...])
        rr_channel = np.sqrt(np.square(xx_channel) + np.square(yy_channel))
        rr_channel = rr_channel / np.amax(rr_channel)

        coords = np.stack([xx_channel, yy_channel, rr_channel], axis=-1)
        return coords

    def reset(self):
        observation = self.env.reset()
        coords = self.generate_coords()
        observation = np.concatenate([observation, coords], axis=-1)
        return observation

    def step(self, action):
        observation, reward, done, info = self.env.step(action)
        coords = self.generate_coords()
        observation = np.concatenate([observation, coords], axis=-1)
        return observation, reward, done, info



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf


class CyclicSchedule(tf.keras.optimizers.schedules.LearningRateSchedule):
    """
    Cyclic learning rate schedule. One cycle during such this
    schedule consists of linearly increasing the learning rate for `step_size`
    steps from `minval` to `maxval`, then linearly decreasing the learning rate
    for another `step_size` steps from `maxval` to (approximately) 0. Note that
    the scheduler is unaware of the total number of planned training steps, so
    be sure to pick `step_size` accordingly.

    Args:
        minval: starting learning rate.
        maxval: maximum learning rate.
        step_size: number of optimization steps in half a cycle.
    """

    def __init__(self, minval, maxval, step_size, decay_to_zero=False):
        self.minval = minval
        self.maxval = maxval
        self.step_size = step_size
        self.decay_to_zero = decay_to_zero

    def get_config(self):
        return {
            "minval": self.minval,
            "maxval": self.maxval,
            "step_size": self.step_size,
        }

    def __call__(self, step):
        step = tf.cast(step, tf.float32)
        cycle = 1 + step // (2 * self.step_size)
        x = step / self.step_size - 2 * cycle + 1
        x_abs = tf.abs(x)
        # decay the downward half towards 0 instead of minval
        if self.decay_to_zero:
            minval = tf.where(x < 0, self.minval, 0.0)
        else:
            minval = self.minval
        vrange = self.maxval - minval
        return minval + vrange * tf.maximum(0.0, 1 - x_abs)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from pyoneer.rl.targets.target_ops import DiscountedReturns, GeneralizedAdvantages


class TargetOpsTest(tf.test.TestCase):
    def test_discounted_returns(self):
        rewards = tf.constant([[0.0, 0.0, 1.0]])
        discounted_rewards = DiscountedReturns(discount_factor=0.99)
        outputs = discounted_rewards(rewards, sample_weight=1.0)
        expected = tf.constant([[0.9801, 0.99, 1.0]])
        self.assertAllClose(outputs, expected)

    def test_generalized_advantages(self):
        rewards = tf.constant([[0.0, 0.0, 1.0]])
        values = tf.constant([[0.0, 0.0, 1.0]])
        generalized_advantages = GeneralizedAdvantages(
            discount_factor=0.99, lambda_factor=0.95, normalize=True
        )
        outputs = generalized_advantages(rewards, values, sample_weight=1.0)
        expected = tf.constant([[0.564769, 0.840459, -1.405228]])
        self.assertAllClose(outputs, expected)


if __name__ == "__main__":
    tf.test.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from pyoneer.math import logical_ops


def safe_divide(x, y, rtol=1e-5, atol=1e-8):
    """
    Safely divide x by y while avoiding dividing by zero.
    """
    y = tf.where(logical_ops.isclose(y, 0.0, rtol=rtol, atol=atol), tf.ones_like(y), y)
    return tf.debugging.check_numerics(x / y, "safe_divide")


def rescale(x, oldmin, oldmax, newmin, newmax):
    """
    Rescale from [oldmin..oldmax] to [newmin..newmax].
    """
    x = tf.convert_to_tensor(x)
    oldmin = tf.convert_to_tensor(oldmin)
    oldmax = tf.convert_to_tensor(oldmax)
    newmin = tf.convert_to_tensor(newmin)
    newmax = tf.convert_to_tensor(newmax)
    x = (x - oldmin) / (oldmax - oldmin)
    x = (x * (newmax - newmin)) + newmin
    x = tf.debugging.check_numerics(x, "rescale")
    return x


def normalize(x, loc, scale, sample_weight=1.0):
    """
    Normalizes an input.

    Args:
        x: a possibly un-normalized Tensor.
        loc: expected loc.
        scale: expected scale.
        sample_weight: optional sample_weight.

    Returns:
        A normalized Tensor.
    """
    x = tf.convert_to_tensor(x)
    loc = tf.convert_to_tensor(loc)
    scale = tf.convert_to_tensor(scale)
    sample_weight = tf.convert_to_tensor(sample_weight)
    outputs = safe_divide((x - loc), scale) * sample_weight
    outputs = tf.debugging.check_numerics(outputs, "normalize")
    return outputs


def denormalize(x, loc, scale, sample_weight=1.0):
    """
    De-normalizes an input.

    Args:
        x: A tensor to denormalize.
        loc: A loc tensor.
        scale: A scale tensor.
        sample_weight: Optional sample_weight tensor.

    Returns:
        A de-normalized Tensor.
    """
    x = tf.convert_to_tensor(x)
    loc = tf.convert_to_tensor(loc)
    scale = tf.convert_to_tensor(scale)
    sample_weight = tf.convert_to_tensor(sample_weight)
    outputs = ((x * scale) + loc) * sample_weight
    outputs = tf.debugging.check_numerics(outputs, "denormalize")
    return outputs



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import tensorflow as tf

from pyoneer.math import angle_ops


class AnglesOpsTest(tf.test.TestCase):
    def test_to_radians(self):
        inputs = tf.constant([-360, -180, -90, 0, +90, +180, +360], dtype=tf.float32)
        outputs = angle_ops.to_radians(inputs)
        expected = tf.constant(
            [
                -math.pi * 2,
                -math.pi,
                -math.pi / 2,
                0,
                +math.pi / 2,
                +math.pi,
                +math.pi * 2,
            ],
            dtype=tf.float32,
        )
        self.assertAllEqual(outputs, expected)

    def test_to_degrees(self):
        inputs = tf.constant(
            [
                -math.pi * 2,
                -math.pi,
                -math.pi / 2,
                0,
                +math.pi / 2,
                +math.pi,
                +math.pi * 2,
            ],
            dtype=tf.float32,
        )
        outputs = angle_ops.to_degrees(inputs)
        expected = tf.constant([-360, -180, -90, 0, +90, +180, +360], dtype=tf.float32)
        self.assertAllEqual(outputs, expected)

    def test_to_cartesian(self):
        inputs = tf.constant(
            [
                -math.pi * 2,
                -math.pi,
                -math.pi / 2,
                0,
                +math.pi / 2,
                +math.pi,
                +math.pi * 2,
            ],
            dtype=tf.float32,
        )
        outputs_x, outputs_y = angle_ops.to_cartesian(inputs)
        expected_x = tf.constant([+1, -1, +0, +1, +0, -1, +1], dtype=tf.float32)
        expected_y = tf.constant([+0, +0, -1, +0, +1, +0, +0], dtype=tf.float32)
        self.assertAllClose(outputs_x, expected_x)
        self.assertAllClose(outputs_y, expected_y)

    def test_to_polar(self):
        inputs_x = tf.constant([+1, -1, +0, +1, +0, -1, +1], dtype=tf.float32)
        inputs_y = tf.constant([+0, +0, -1, +0, +1, +0, +0], dtype=tf.float32)
        outputs_rho, outputs_phi = angle_ops.to_polar(inputs_x, inputs_y)
        expected_rho = tf.constant([1, 1, 1, 1, 1, 1, 1], dtype=tf.float32)
        expected_phi = tf.constant(
            [0, math.pi, -math.pi / 2, 0, math.pi / 2, math.pi, 0], dtype=tf.float32
        )
        self.assertAllClose(outputs_rho, expected_rho)
        self.assertAllClose(outputs_phi, expected_phi)


if __name__ == "__main__":
    tf.test.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc

import six
import tensorflow as tf

from pyoneer.math import math_ops


def range_moments(minval, maxval):
    """
    Compute element-wise mean and variance from min and max values.

    Args:
        minval: A tensor of minimum values.
        maxval: A tensor of maximum values.

    Returns:
        Tuple of (mean, variance).
    """
    mean = (maxval + minval) / 2
    variance = tf.square((maxval - minval) / 2)
    return mean, variance


@six.add_metaclass(abc.ABCMeta)
class Moments(tf.keras.layers.Layer):
    """
    Base class for moment containers.
    """

    @property
    @abc.abstractmethod
    def mean(self):
        raise NotImplementedError("Must be implemented in subclasses.")

    @property
    @abc.abstractmethod
    def variance(self):
        raise NotImplementedError("Must be implemented in subclasses.")

    @property
    def std(self):
        return tf.sqrt(self.variance)

    @abc.abstractmethod
    def update_state(self, *args, **kwargs):
        raise NotImplementedError("Must be implemented in subclasses.")

    def normalize(self, inputs, sample_weight=1.0):
        return math_ops.normalize(
            inputs, loc=self.mean, scale=self.std, sample_weight=sample_weight
        )

    def denormalize(self, inputs, sample_weight=1.0):
        return math_ops.denormalize(
            inputs, loc=self.mean, scale=self.std, sample_weight=sample_weight
        )


class StaticMoments(Moments):
    """
    Static moments.

    Args:
        mean: Mean of moments.
        variance: Variance of moments.
    """

    def __init__(self, mean, variance, **kwargs):
        super(StaticMoments, self).__init__(**kwargs)
        self._mean = tf.Variable(mean, trainable=False)
        self._variance = tf.Variable(variance, trainable=False)

    @property
    def mean(self):
        return self._mean

    @property
    def variance(self):
        return self._variance

    def update_state(self, *args, **kwargs):
        raise NotImplementedError("StaticMoments cannot be updated.")


class StreamingMoments(Moments):
    """
    Compute accurate moments on a batch basis.

    Args:
        shape: Shape of the moments. Used to infer the reduction axes.
        dtype: Optional dtype.
    """

    def __init__(self, shape, dtype=tf.float32, **kwargs):
        super(StreamingMoments, self).__init__(**kwargs)
        self._count = self.add_weight(
            name="count",
            initializer="zeros",
            shape=shape,
            dtype=tf.int64,
            trainable=False,
        )
        self._mean = self.add_weight(
            name="mean", initializer="zeros", shape=shape, dtype=dtype, trainable=False
        )
        self._var_sum = self.add_weight(
            name="var_sum",
            initializer="zeros",
            shape=shape,
            dtype=dtype,
            trainable=False,
        )

    @property
    def count(self):
        return self._count

    @property
    def mean(self):
        return self._mean

    @property
    def variance(self):
        return tf.where(
            self._count > 1,
            self._var_sum / tf.cast(self._count - 1, self._var_sum.dtype),
            tf.zeros_like(self._var_sum),
        )

    def update_state(self, inputs, sample_weight=None):
        """
        Update moments using a new batch of inputs.

        Args:
            inputs: Input tensor.
            sample_weight: Optional sample weight.
        """
        inputs = tf.convert_to_tensor(inputs)

        if sample_weight is None:
            sample_weight = 1.0

        ndims = inputs.shape.ndims - self._mean.shape.ndims
        axes = list(range(ndims))

        weight_sum = tf.reduce_sum(sample_weight, axis=axes)
        count_delta = tf.cast(weight_sum, tf.int64)
        new_count = self._count + count_delta

        mean_delta = tf.where(
            count_delta > 1,
            (
                tf.reduce_sum((inputs - self._mean) * sample_weight, axis=axes)
                / tf.cast(new_count, tf.float32)
            ),
            math_ops.safe_divide(
                tf.reduce_sum(inputs * sample_weight, axis=axes), weight_sum
            ),
        )
        new_mean = self._mean + mean_delta

        var_delta = tf.reduce_sum(
            (inputs - self._mean) * (inputs - new_mean) * sample_weight, axis=axes
        )
        new_var_sum = self._var_sum + var_delta

        self._count.assign(new_count)
        self._mean.assign(new_mean)
        self._var_sum.assign(new_var_sum)


class ExponentialMovingMoments(Moments):
    """
    Compute moments as an exponential moving average using the update rule:

    ```
    mean = (1 - rate) * old_mean + rate * new_mean
    variance = (1 - rate) * old_variance + rate * new_variance
    ```

    Args:
        rate: Update rate in the range [0, 1] of the exponential moving
            average. Smaller values update faster.
        shape: Shape of the moments. Used to infer the reduction axes.
        dtype: Optional dtype.
    """

    def __init__(self, rate, shape, dtype=tf.float32, **kwargs):
        super(ExponentialMovingMoments, self).__init__(**kwargs)
        self._count = self.add_weight(
            name="count",
            initializer="zeros",
            shape=shape,
            dtype=tf.int64,
            trainable=False,
        )
        self._mean = self.add_weight(
            name="mean", initializer="zeros", shape=shape, dtype=dtype, trainable=False
        )
        self._variance = self.add_weight(
            name="variance",
            initializer="zeros",
            shape=shape,
            dtype=dtype,
            trainable=False,
        )
        self.rate = rate

    @property
    def count(self):
        return self._count

    @property
    def mean(self):
        return self._mean

    @property
    def variance(self):
        return self._variance

    def update_state(self, inputs, sample_weight=None):
        """
        Update moments using a new batch of inputs.

        Args:
            inputs: Input tensor.
            sample_weight: Optional sample weight.
        """
        inputs = tf.convert_to_tensor(inputs)

        if sample_weight is None:
            sample_weight = 1.0

        ndims = inputs.shape.ndims - self._mean.shape.ndims
        axes = list(range(ndims))

        weight_sum = tf.reduce_sum(sample_weight, axis=axes)
        count = tf.cast(weight_sum, tf.int64)
        new_count = self._count + count

        mean, variance = tf.nn.weighted_moments(
            inputs, axes=axes, frequency_weights=sample_weight
        )

        moving_mean = self._mean * self.rate + mean * (1 - self.rate)
        moving_variance = self._variance * self.rate + variance * (1 - self.rate)

        new_mean = tf.where(self._count > 0, moving_mean, mean)
        new_variance = tf.where(self._count > 0, moving_variance, variance)

        new_mean = tf.where(count > 0, new_mean, self._mean)
        new_variance = tf.where(count > 0, new_variance, self._variance)

        self._count.assign(new_count)
        self._mean.assign(new_mean)
        self._variance.assign(new_variance)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from pyoneer.math import math_ops


class MathOpsTest(tf.test.TestCase):
    def test_safe_divide(self):
        x = tf.ones(shape=[7], dtype=tf.float32)
        y = tf.constant([-1.0, -0.5, -0.2, 0.0, +0.2, +0.5, +1.0])
        actual = math_ops.safe_divide(x, y)
        expected = tf.constant([-1.0, -2.0, -5.0, 1.0, +5.0, +2.0, +1.0])
        self.assertAllEqual(actual, expected)

    def test_rescale(self):
        x = tf.constant([0.5, 0.5, 0.5])
        actual = math_ops.rescale(x, 0.0, 1.0, -1.0, 1.0)
        expected = tf.zeros(shape=[3], dtype=tf.float32)
        self.assertAllEqual(actual, expected)

    def test_normalize(self):
        x = tf.constant([-1.0, 0.0, 1.0, 0.0])
        sample_weight = tf.constant([1.0, 1.0, 1.0, 0.0])
        actual = math_ops.normalize(x, loc=-1.0, scale=2.0, sample_weight=sample_weight)
        expected = tf.constant([0.0, 0.5, 1.0, 0.0])
        self.assertAllClose(actual, expected)

    def test_denormalize(self):
        x = tf.constant([0.0, 0.5, 1.0, 0.0])
        sample_weight = tf.constant([1.0, 1.0, 1.0, 0.0])
        actual = math_ops.denormalize(
            x, loc=-1.0, scale=2.0, sample_weight=sample_weight
        )
        expected = tf.constant([-1.0, 0.0, 1.0, 0.0])
        self.assertAllClose(actual, expected)


if __name__ == "__main__":
    tf.test.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from pyoneer.manip.array_ops import flatten, pad_or_truncate, shift
from pyoneer.manip.indexing_ops import batched_index

__all__ = ["flatten", "pad_or_truncate", "shift", "batched_index"]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow.python.keras.utils import losses_utils

from pyoneer.rl.losses.policy_gradient_ops import (
    PolicyGradient,
    ClippedPolicyGradient,
    PolicyEntropy,
)


class PolicyGradientTest(tf.test.TestCase):
    def test_policy_gradient(self):
        log_probs = tf.math.log(tf.constant([0.9, 0.8, 0.8, 0.8]))
        advantages = tf.constant([1.0, 0.0, 1.0, 0.0])
        sample_weight = tf.constant([1.0, 0.0, 1.0, 0.0])

        loss_fn = PolicyGradient()
        loss = loss_fn(
            log_probs=log_probs, advantages=advantages, sample_weight=sample_weight
        )
        expected = losses_utils.compute_weighted_loss(
            -log_probs * advantages, sample_weight=sample_weight
        )

        self.assertAllClose(loss, expected)

    def test_clipped_policy_gradient(self):
        epsilon_clipping = 0.2

        log_probs = tf.math.log(tf.constant([0.9, 0.8, 0.8, 0.8]))
        log_probs_anchor = tf.math.log(tf.constant([0.95, 0.85, 0.85, 0.85]))

        advantages = tf.constant([1.0, 0.0, 1.0, 0.0])
        sample_weight = tf.constant([1.0, 0.0, 1.0, 0.0])

        # using the function
        loss_fn = ClippedPolicyGradient(epsilon_clipping=epsilon_clipping)
        loss = loss_fn(
            log_probs=log_probs,
            log_probs_anchor=log_probs_anchor,
            advantages=advantages,
            sample_weight=sample_weight,
        )

        # manual check
        ratio = tf.exp(log_probs - log_probs_anchor)
        surrogate1 = ratio * advantages
        surrogate2 = (
            tf.clip_by_value(ratio, 1 - epsilon_clipping, 1 + epsilon_clipping)
            * advantages
        )

        surrogate_min = tf.minimum(surrogate1, surrogate2)

        expected = -losses_utils.compute_weighted_loss(
            losses=surrogate_min, sample_weight=sample_weight
        )

        self.assertAllClose(loss, expected)

    def test_entropy(self):
        entropy = tf.constant([0.2, 0.3, 0.4])
        expected = -tf.constant(0.3)
        loss_fn = PolicyEntropy()
        losses = loss_fn(entropy)
        self.assertAllClose(losses, expected)


if __name__ == "__main__":
    tf.test.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from pyoneer.rl.targets.target_ops import DiscountedReturns, GeneralizedAdvantages

__all__ = ["DiscountedReturns", "GeneralizedAdvantages"]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow.python.keras.metrics import MeanMetricWrapper


def mape(y_true, y_pred):
    """
    Calculates the mean absolute percentage error of the predicted values to
    the true values. `y_true`, `y_pred` should be of the same type and
    shape.

    Args:
        y_true: Tensor of ground-truth values.
        y_pred: Tensor of predicted values.
    """
    y_pred.get_shape().assert_is_compatible_with(y_true.get_shape())
    if y_true.dtype != y_pred.dtype:
        y_pred = tf.cast(y_pred, y_true.dtype)
    errors = tf.abs((y_pred - y_true) / y_true)
    errors = tf.reduce_mean(errors, axis=-1)
    return errors


def smape(y_true, y_pred):
    """
    Calculates the symmetric mean absolute percentage error of the predicted
    values to the true values. `y_true`, `y_pred` should be of the same
    type and shape.

    The SMAPE is calculated as:

        2 * mean(|y_pred - y_true| / (|y_pred| + |y_true|))

    Therefore, it is bounded in the range of [0, 2].

    Args:
        y_true: Tensor of ground-truth values.
        y_pred: Tensor of predicted values.
    """
    y_pred.get_shape().assert_is_compatible_with(y_true.get_shape())
    if y_true.dtype != y_pred.dtype:
        y_pred = tf.cast(y_pred, y_true.dtype)
    errors = 2 * tf.abs(y_pred - y_true) / (tf.abs(y_pred) + tf.abs(y_true))
    errors = tf.reduce_mean(errors, axis=-1)
    return errors


class MAPE(MeanMetricWrapper):
    """
    Calculates the mean absolute percentage error of predicted values to the
    actual ground-truth values.

    Attributes:
        name: name of the MAPE object.
        dtype: data type of the tensor.
    """

    def __init__(self, name="MAPE", dtype=None):
        """Inits MAPE class with name and dtype."""
        super(MAPE, self).__init__(mape, name, dtype=dtype)


class SMAPE(MeanMetricWrapper):
    """
    Calculates the symmetric mean absolute percentage error of predicted
    values to the actual ground-truth values. The SMAPE is calculated as:

        2 * mean(|y_pred - y_true| / (|y_pred| + |y_true|))

    Therefore, it's bounded in the range of [0, 2].

    Attributes:
        name: name of the SMAPE object.
        dtype: data type of the tensor.
    """

    def __init__(self, name="SMAPE", dtype=None):
        """Inits SMAPE class with name and dtype."""
        super(SMAPE, self).__init__(smape, name, dtype=dtype)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf
import tensorflow_probability as tfp

from pyoneer.rl.strategies import strategies_impl


class StrategiesTest(tf.test.TestCase):
    def test_epsilon_greedy(self):
        logits = [[0.0, 1.0, -2.0], [0.0, 1.0, 2.0]]

        def policy(x):
            return tfp.distributions.Categorical(logits=x)

        expected_samples = tf.constant([1, 2])
        expected_policy_samples = policy(logits).mode()
        self.assertAllEqual(expected_policy_samples, expected_samples)

        epsilon_greedy = strategies_impl.EpsilonGreedy(policy, 0.0)
        samples = epsilon_greedy(logits)
        self.assertAllEqual(samples, expected_policy_samples)

    def test_sample(self):
        logits = [[0.0, 1.0, -2.0], [0.0, 1.0, 2.0]]

        def policy(x):
            return tfp.distributions.Categorical(logits=x)

        expected_samples = tf.constant([1, 2])
        expected_policy_samples = policy(logits).mode()
        self.assertAllEqual(expected_policy_samples, expected_samples)

        sampler = strategies_impl.Sample(policy)
        samples = sampler(logits)
        self.assertShapeEqual(samples.numpy(), expected_policy_samples)

    def test_mode(self):
        logits = [[0.0, 1.0, -2.0], [0.0, 1.0, 2.0]]

        def policy(x):
            return tfp.distributions.Categorical(logits=x)

        expected_samples = tf.constant([1, 2])
        expected_policy_samples = policy(logits).mode()
        self.assertAllEqual(expected_policy_samples, expected_samples)

        sampler = strategies_impl.Mode(policy)
        samples = sampler(logits)
        self.assertShapeEqual(samples.numpy(), expected_policy_samples)


if __name__ == "__main__":
    tf.test.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from pyoneer.schedules.cyclic_schedule_impl import CyclicSchedule

__all__ = ["CyclicSchedule"]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import gym
import numpy as np
import tensorflow as tf

from pyoneer.rl.wrappers.observation_impl import (
    ObservationCoordinates,
    ObservationNormalization,
)


class TestEnv(gym.Env):
    def __init__(self):
        self.observation_space = gym.spaces.Box(
            low=0.0, high=1.0, shape=(4, 4, 1), dtype=np.float32
        )
        self.action_space = gym.spaces.Box(
            low=0.0, high=1.0, shape=(), dtype=np.float32
        )

    def reset(self):
        return self.observation_space.sample()

    def step(self, action):
        state = self.observation_space.sample()
        reward = 0.0
        done = False
        info = {}
        return state, reward, done, info


class ObservationTest(tf.test.TestCase):
    def test_observation_coords(self):
        env = TestEnv()
        env = ObservationCoordinates(env)
        self.assertTupleEqual(env.observation_space.shape, (4, 4, 4))

    def test_observation_norm(self):
        env = TestEnv()
        env = ObservationNormalization(env)
        self.assertTupleEqual(env.observation_space.shape, (4, 4, 1))


if __name__ == "__main__":
    tf.test.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf
import tensorflow_probability as tfp


class SoftplusInverse(tf.keras.initializers.Initializer):
    """
    Initializer that generates tensors initialized to `log(exp(scale) - 1)`.

    Args:
        scale: scale of the initializer output.
        dtype: dtype of the operation.
    """

    def __init__(self, scale=1.0):
        self.scale = scale

    def __call__(self, shape, dtype=tf.float32):
        return tf.constant(
            tfp.math.softplus_inverse(self.scale), dtype=dtype, shape=shape
        )

    def get_config(self):
        return {"scale": self.scale, "dtype": self.dtype.name}



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from pyoneer.rl import losses
from pyoneer.rl import strategies
from pyoneer.rl import targets
from pyoneer.rl import wrappers

__all__ = ["losses", "strategies", "targets", "wrappers"]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf


def batched_index(values, indices):
    """
    Performs indexing on batches and sequence-batches by reducing over
    zero-masked values.

    Equivalent to `values[:, indices]`.

    Args:
        values: tensor of shape `[B, num_values]` or `[T, B, num_values]`
        indices: tensor of shape `[B]` or `[T, B]` containing indices.
    Returns:
        Tensor of shape `[B]` or `[T, B]` containing values for the given
        indices.
    """
    values = tf.convert_to_tensor(values)
    indices = tf.convert_to_tensor(indices)
    mask = tf.one_hot(indices, values.shape[-1], dtype=values.dtype)
    return tf.reduce_sum(values * mask, axis=-1)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from pyoneer.rl.wrappers.observation_impl import (
    ObservationCoordinates,
    ObservationNormalization,
)
from pyoneer.rl.wrappers.batch_impl import Batch
from pyoneer.rl.wrappers.process_impl import Process

__all__ = ["ObservationCoordinates", "ObservationNormalization", "Batch", "Process"]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
import atexit
import traceback
import multiprocessing


class Process(object):
    """
    Wraps a `gym.Env` to host the environment in an external process.

    Example:

    ```
    env = Process(lambda: gym.make('Pendulum-v0'))
    ```

    Args:
        constructor: Constructor which returns a `gym.Env`.
    """

    _ACCESS = 1
    _CALL = 2
    _RESULT = 3
    _EXCEPTION = 4
    _CLOSE = 5

    def __init__(self, constructor):
        self._conn, conn = multiprocessing.Pipe()
        self._process = multiprocessing.Process(
            target=self._worker, args=(constructor, conn)
        )

        atexit.register(self.close)

        self._process.start()
        self._observation_space = None
        self._action_space = None

    @property
    def observation_space(self):
        if self._observation_space is None:
            self._observation_space = self.__getattr__("observation_space")
        return self._observation_space

    @property
    def action_space(self):
        if self._action_space is None:
            self._action_space = self.__getattr__("action_space")
        return self._action_space

    def __getattr__(self, name):
        self._conn.send((self._ACCESS, name))
        return self._receive()

    def call(self, name, *args, **kwargs):
        payload = name, args, kwargs
        self._conn.send((self._CALL, payload))
        return self._receive

    def close(self):
        try:
            self._conn.send((self._CLOSE, None))
            self._conn.close()
        except IOError:
            # connection already closed
            pass

        self._process.join()

    def seed(self, seed):
        return self.call("seed", seed)

    def step(self, action):
        return self.call("step", action)

    def reset(self):
        return self.call("reset")

    def _receive(self):
        message, payload = self._conn.recv()

        # re-raise exceptions in the main process
        if message == self._EXCEPTION:
            stacktrace = payload
            raise Exception(stacktrace)

        if message == self._RESULT:
            return payload

        raise KeyError("Received message of unexpected type {}".format(message))

    def _worker(self, constructor, conn):
        try:
            env = constructor()

            while True:
                try:
                    # only block for short times to support keyboard exceptions
                    if not conn.poll(0.1):
                        continue
                    message, payload = conn.recv()
                except (EOFError, KeyboardInterrupt):
                    break

                if message == self._ACCESS:
                    name = payload
                    result = getattr(env, name)
                    conn.send((self._RESULT, result))
                    continue

                if message == self._CALL:
                    name, args, kwargs = payload
                    result = getattr(env, name)(*args, **kwargs)
                    conn.send((self._RESULT, result))
                    continue

                if message == self._CLOSE:
                    assert payload is None
                    break

                raise KeyError("Received message of unknown type {}".format(message))

        except Exception:
            stacktrace = "".join(traceback.format_exception(*sys.exc_info()))
            conn.send((self._EXCEPTION, stacktrace))
        finally:
            conn.close()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import gym
import tensorflow as tf

from pyoneer.rl.wrappers.process_impl import Process


class ProcessTest(tf.test.TestCase):
    def test_process(self):
        env = Process(lambda: gym.make("Pendulum-v0"))

        promise = env.seed(0)
        promise()

        promise = env.reset()
        state = promise()

        action = env.action_space.sample()
        promise = env.step(action)
        next_state, reward, done, info = promise()

        self.assertTupleEqual(state.shape, (3,))
        self.assertTupleEqual(action.shape, (1,))
        self.assertTupleEqual(next_state.shape, (3,))


if __name__ == "__main__":
    tf.test.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from pyoneer.manip import indexing_ops


class IndexingOpsTest(tf.test.TestCase):
    def test_batched_index(self):
        values = tf.constant([[0, 1, 2], [0, 1, 2], [0, 1, 2]])
        indices = tf.constant([0, 1, 2])
        output = indexing_ops.batched_index(values, indices)
        expected = tf.constant([0, 1, 2])
        self.assertAllEqual(output, expected)


if __name__ == "__main__":
    tf.test.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from pyoneer.activations import activations_impl
from pyoneer.math import angle_ops


class Swish(tf.keras.layers.Layer):
    """
    Compute Swish, self-gating, activation function layer: `x * sigmoid(x)`.
    """

    def call(self, inputs):
        """
        Compute activations for the inputs.

        Args:
            inputs: A tensor.

        Returns:
            The activations.
        """
        outputs = activations_impl.swish(inputs)
        return outputs


class OneHotEncoder(tf.keras.layers.Layer):
    """
    One-hot encoding layer. Encodes the integer inputs as one-hot vectors.

    Args:
        depth: The depth of the one-hot encoding.
    """

    def __init__(self, depth, **kwargs):
        super(OneHotEncoder, self).__init__(**kwargs)
        self.depth = depth

    def call(self, inputs):
        """
        Encode the inputs.

        Args:
            inputs: An integer tensor.

        Returns:
            The one-hot encoded inputs.
        """
        inputs = tf.cast(inputs, tf.int64)
        outputs = tf.one_hot(inputs, self.depth)
        outputs = tf.debugging.check_numerics(outputs, "outputs")
        return outputs

    def get_config(self):
        config = {"depth": self.depth}
        base_config = super(OneHotEncoder, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class AngleEncoder(tf.keras.layers.Layer):
    """
    Angle encoding layer. Encodes an angle as the cosine and sine of radians
    or degrees which will be converted to radians.

    Args:
        degrees (default: False):
            Whether the inputs are in degrees (True) or radians (False).
    """

    def __init__(self, degrees=False, **kwargs):
        super(AngleEncoder, self).__init__(**kwargs)
        self.degrees = degrees

    def call(self, inputs):
        if self.degrees:
            inputs = angle_ops.to_radians(inputs)
        x, y = angle_ops.to_cartesian(inputs)
        outputs = tf.concat([x, y], axis=-1)
        return outputs

    def get_config(self):
        config = {"degrees": self.degrees}
        base_config = super(AngleEncoder, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class Nest(tf.keras.layers.Layer):
    """
    Maps an arbitrarily nested structure of layers to an equivalent
    structure of input features using the `tf.nest` API.

    Example:

    ```
    nest_layer = Nest({
        "category": OneHotEncoder(depth=3),
        "angle": AngleEncoder(degrees=True),
        "nested": {
            "category": OneHotEncoder(depth=3),
            "angle": AngleEncoder(degrees=True),
        }
    })
    features = {
        "category": tf.constant([1]),
        "angle": tf.constant([[-45.0]]),
        "nested": {
            "category": tf.constant([2]),
            "angle": tf.constant([[+45.0]]),
        }
    }
    outputs = nest_layer(features)  # shape: [1, (2 + 3) * 2]
    ```

    Args:
        layers: An arbitrarily nested structure of layers.
    """

    def __init__(self, layers, **kwargs):
        super(Nest, self).__init__(**kwargs)
        self.layers = layers

    def call(self, inputs):
        tf.nest.assert_same_structure(self.layers, inputs)

        def call_layer(layer, inputs):
            return layer(inputs)

        inputs_mapped = tf.nest.map_structure(call_layer, self.layers, inputs)
        inputs_flattened = tf.nest.flatten(inputs_mapped)
        outputs = tf.concat(inputs_flattened, axis=-1)
        return outputs



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from pyoneer.moments.moments_impl import (
    range_moments,
    StaticMoments,
    StreamingMoments,
    ExponentialMovingMoments,
)

__all__ = [
    "range_moments",
    "StaticMoments",
    "StreamingMoments",
    "ExponentialMovingMoments",
]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from pyoneer.metrics.metrics_impl import mape, smape, MAPE, SMAPE

__all__ = ["mape", "smape", "MAPE", "SMAPE"]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from pyoneer.math.angle_ops import (
    to_radians,
    to_degrees,
    to_cartesian,
    to_polar,
    RADIANS_TO_DEGREES,
    DEGREES_TO_RADIANS,
)
from pyoneer.math.logical_ops import isclose
from pyoneer.math.math_ops import safe_divide, rescale, normalize, denormalize

__all__ = [
    "to_radians",
    "to_degrees",
    "to_cartesian",
    "to_polar",
    "RADIANS_TO_DEGREES",
    "DEGREES_TO_RADIANS",
    "isclose",
    "safe_divide",
    "rescale",
    "normalize",
    "denormalize",
]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from pyoneer.initializers.initializers_impl import SoftplusInverse


class InitializersTest(tf.test.TestCase):
    def test_softplus_inverse(self):
        initializer = SoftplusInverse(1.0)
        output = initializer(shape=[1, 3])
        expected = tf.constant([[0.54132485, 0.54132485, 0.54132485]], dtype=tf.float32)
        self.assertAllEqual(output, expected)


if __name__ == "__main__":
    tf.test.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from pyoneer.moments import moments_impl


class MomentsTest(tf.test.TestCase):
    def test_range_moments(self):
        mean, variance = moments_impl.range_moments(
            minval=tf.constant([-2.0]), maxval=tf.constant([2.0])
        )
        expected_mean = tf.constant([0.0], dtype=tf.float32)
        expected_variance = tf.constant([4.0], dtype=tf.float32)
        self.assertAllEqual(mean, expected_mean)
        self.assertAllEqual(variance, expected_variance)

    def test_static_moments(self):
        mean = tf.constant([-1.0, 0.0, +1.0], dtype=tf.float32)
        variance = tf.constant([1.0, 2.0, 1.0], dtype=tf.float32)
        moments = moments_impl.StaticMoments(mean, variance)

        expected_mean = tf.constant([-1.0, 0.0, +1.0], dtype=tf.float32)
        expected_var = tf.constant([1.0, 2.0, 1.0], dtype=tf.float32)
        expected_std = tf.sqrt(expected_var)

        self.assertAllClose(expected_mean, moments.mean.numpy())
        self.assertAllClose(expected_var, moments.variance.numpy())
        self.assertAllClose(expected_std, moments.std.numpy())

    def test_streaming_moments(self):
        moments = moments_impl.StreamingMoments(shape=[3])

        # sample 1
        inputs = tf.constant([[[-1.0, 0.0, +1.0]]])
        sample_weight = tf.constant([[[1.0, 0.0, 1.0]]])

        moments.update_state(inputs, sample_weight=sample_weight)

        expected_mean = tf.constant([-1.0, 0.0, +1.0], dtype=tf.float32)
        expected_var = tf.constant([0.0, 0.0, 0.0], dtype=tf.float32)
        expected_std = tf.sqrt(expected_var)

        self.assertAllClose(expected_mean, moments.mean.numpy())
        self.assertAllClose(expected_var, moments.variance.numpy())
        self.assertAllClose(expected_std, moments.std.numpy())

        # sample 2
        inputs = tf.constant([[[0.0, +1.0, -1.0]]])
        sample_weight = tf.constant([[[0.0, 1.0, 1.0]]])

        moments.update_state(inputs, sample_weight=sample_weight)

        expected_mean = tf.constant([-1.0, +1.0, 0.0], dtype=tf.float32)
        expected_var = tf.constant([0.0, 0.0, 2.0], dtype=tf.float32)
        expected_std = tf.sqrt(expected_var)

        self.assertAllClose(expected_mean, moments.mean.numpy())
        self.assertAllClose(expected_var, moments.variance.numpy())
        self.assertAllClose(expected_std, moments.std.numpy())

        # sample 3
        inputs = tf.constant([[[+1.0, -1.0, 0.0]]])
        sample_weight = tf.constant([[[1.0, 1.0, 0.0]]])

        moments.update_state(inputs, sample_weight=sample_weight)

        expected_mean = tf.constant([0.0, 0.0, 0.0], dtype=tf.float32)
        expected_var = tf.constant([2.0, 2.0, 2.0], dtype=tf.float32)
        expected_std = tf.sqrt(expected_var)

        self.assertAllClose(expected_mean, moments.mean.numpy())
        self.assertAllClose(expected_var, moments.variance.numpy())
        self.assertAllClose(expected_std, moments.std.numpy())

    def test_exponential_moving_moments(self):
        moments = moments_impl.ExponentialMovingMoments(shape=[3], rate=0.9)

        # sample 1
        inputs = tf.constant([[[-1.0, 0.0, +1.0]]])
        sample_weight = tf.constant([[[1.0, 0.0, 1.0]]])

        moments.update_state(inputs, sample_weight=sample_weight)

        expected_mean = tf.constant([-1.0, 0.0, +1.0], dtype=tf.float32)
        expected_var = tf.constant([0.0, 0.0, 0.0], dtype=tf.float32)
        expected_std = tf.sqrt(expected_var)

        self.assertAllClose(expected_mean, moments.mean.numpy())
        self.assertAllClose(expected_var, moments.variance.numpy())
        self.assertAllClose(expected_std, moments.std.numpy())

        # sample 2
        inputs = tf.constant([[[0.0, +1.0, -1.0]]])
        sample_weight = tf.constant([[[0.0, 1.0, 1.0]]])

        moments.update_state(inputs, sample_weight=sample_weight)

        expected_mean = tf.constant([-1.0, 1.0, 0.8], dtype=tf.float32)
        expected_var = tf.constant([0.0, 0.0, 0.0], dtype=tf.float32)
        expected_std = tf.sqrt(expected_var)

        self.assertAllClose(expected_mean, moments.mean.numpy())
        self.assertAllClose(expected_var, moments.variance.numpy())
        self.assertAllClose(expected_std, moments.std.numpy())

        # sample 3
        inputs = tf.constant([[[+1.0, -1.0, 0.0]]])
        sample_weight = tf.constant([[[1.0, 1.0, 0.0]]])

        moments.update_state(inputs, sample_weight=sample_weight)

        expected_mean = tf.constant([-0.8, 0.8, 0.8], dtype=tf.float32)
        expected_var = tf.constant([0.0, 0.0, 0.0], dtype=tf.float32)
        expected_std = tf.sqrt(expected_var)

        self.assertAllClose(expected_mean, moments.mean.numpy())
        self.assertAllClose(expected_var, moments.variance.numpy())
        self.assertAllClose(expected_std, moments.std.numpy())


if __name__ == "__main__":
    tf.test.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from pyoneer.activations.activations_impl import swish


class ActivationsTest(tf.test.TestCase):
    def test_swish(self):
        x = tf.constant([-1.0, 0.0, +1.0])
        actual = swish(x)
        expected = tf.constant([-0.268941, 0.0, 0.731059])
        self.assertAllClose(actual, expected)


if __name__ == "__main__":
    tf.test.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

from pyoneer.rl.wrappers.process_impl import Process


class Batch(object):
    """
    Wraps multiple `gym.Env` into a batch of environments to support
    vectorization. Uses an external processes for each environment
    when `blocking=False`.

    For simple environments, use `blocking=True` as the overhead of
    external processes can limit very large batches.

    For more complex environments, use `blocking=False` so the CPU
    computation can be offloaded.

    Note: When rendering with `mode="human"` only the first
        environment is rendered.

    Example:

    ```
    env = Batch(
        constructor=lambda: gym.make('Pendulum-v0'),
        batch_size=32,
        blocking=True)
    ```

    Args:
        constructor: Constructor which returns a `gym.Env`.
        batch_size: Number of parallel environments.
        blocking: Boolean indicating whether each call to an environment
            is blocking (default: True).
    """

    def __init__(self, constructor, batch_size, blocking=True):
        if blocking:
            self.envs = [constructor() for _ in range(batch_size)]
        else:
            self.envs = [Process(constructor) for _ in range(batch_size)]

        self.done = np.zeros(len(self.envs), dtype=np.bool)
        self.blocking = blocking

        observation_space = self.observation_space
        if not all(env.observation_space == observation_space for env in self.envs):
            raise ValueError("All environments must use the same observation space.")

        action_space = self.action_space
        if not all(env.action_space == action_space for env in self.envs):
            raise ValueError("All environments must use the same action space.")

    def __len__(self):
        return len(self.envs)

    def __getitem__(self, index):
        return self.envs[index]

    def __getattr__(self, name):
        return getattr(self.envs[0], name)

    def seed(self, seed):
        if self.blocking:
            for i, env in enumerate(self.envs):
                env.seed(seed + i)
        else:
            promises = [env.seed(seed + i) for i, env in enumerate(self.envs)]
            for promise in promises:
                promise()

    def reset(self):
        self.done[:] = False

        if self.blocking:
            states = [env.reset() for env in self.envs]
        else:
            promises = [env.reset() for env in self.envs]
            states = [promise() for promise in promises]

        state = np.stack(states, axis=0)
        return state

    def _dummy_transition(self):
        next_state = np.zeros(
            shape=self.observation_space.shape, dtype=self.observation_space.dtype
        )
        reward = 0.0
        done = True
        info = {}
        transition = (next_state, reward, done, info)
        return transition

    def step(self, actions):
        if self.blocking:
            transitions = []
            for i, env in enumerate(self.envs):
                if self.done[i]:
                    transition = self._dummy_transition()
                else:
                    transition = env.step(actions[i])
                transitions.append(transition)
        else:
            promises = []
            for i, env in enumerate(self.envs):
                if self.done[i]:
                    promise = self._dummy_transition
                else:
                    promise = env.step(actions[i])
                promises.append(promise)
            transitions = [promise() for promise in promises]

        next_states, rewards, dones, infos = zip(*transitions)
        next_state = np.stack(next_states, axis=0)
        reward = np.stack(rewards, axis=0)
        done = np.stack(dones, axis=0)
        info = tuple(infos)
        self.done = self.done | done
        return next_state, reward, done, info

    def render(self, mode="human"):
        assert (
            self.blocking or mode == "rgb_array"
        ), 'only the "rgb_array" mode is supported when `blocking=False`'

        if mode == "rgb_array":
            return np.stack([env.render(mode=mode) for env in self.envs], axis=0)
        else:
            return self.envs[0].render(mode=mode)

    def close(self):
        for env in self.envs:
            if hasattr(env, "close"):
                env.close()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from pyoneer import metrics


class MetricsTest(tf.test.TestCase):
    def test_mape_fn(self):
        y_true = tf.constant([[0.2, 0.1], [0.3, 0.2], [0.1, 0.2]], dtype=tf.float32)
        y_pred = tf.constant([[0.1, 0.1], [0.2, 0.1], [0.2, 0.2]], dtype=tf.float32)

        result = metrics.mape(y_true, y_pred)
        expected = tf.constant([0.25, 0.416667, 0.5], dtype=tf.float32)
        self.assertAllClose(result, expected)

    def test_smape_fn(self):
        y_true = tf.constant([[0.2, 0.1], [0.3, 0.2], [0.1, 0.2]], dtype=tf.float32)
        y_pred = tf.constant([[0.1, 0.1], [0.2, 0.1], [0.2, 0.2]], dtype=tf.float32)

        result = metrics.smape(y_true, y_pred)
        expected = tf.constant([0.333333, 0.533333, 0.333333], dtype=tf.float32)
        self.assertAllClose(result, expected)

    def test_mape(self):
        y_true = tf.constant([[0.2, 0.1], [0.3, 0.2], [0.1, 0.2]], dtype=tf.float32)
        y_pred = tf.constant([[0.1, 0.1], [0.2, 0.1], [0.2, 0.2]], dtype=tf.float32)
        sample_weight = tf.constant([1.0, 0.0, 1.0], dtype=tf.float32)

        metric = metrics.MAPE()
        metric.update_state(y_true, y_pred, sample_weight=sample_weight)
        self.assertAllClose(metric.result(), 0.375)

    def test_smape(self):
        y_true = tf.constant([[0.3, 0.1], [0.3, 0.3], [0.1, 0.2]], dtype=tf.float32)
        y_pred = tf.constant([[0.1, 0.1], [0.2, 0.1], [0.3, 0.2]], dtype=tf.float32)
        sample_weight = tf.constant([1.0, 0.0, 1.0], dtype=tf.float32)

        metric = metrics.SMAPE()
        metric.update_state(y_true, y_pred, sample_weight=sample_weight)
        self.assertAllClose(metric.result(), 0.5)


if __name__ == "__main__":
    tf.test.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow.python.keras.utils import losses_utils


def policy_gradient(log_probs, advantages):
    """
    Computes the Vanilla policy gradient loss.

    Args:
        log_probs: Log probabilities of taking actions under a policy.
        advantages: Advantage estimation.

    Returns:
        Tensor of losses.
    """
    advantages = tf.stop_gradient(advantages)
    losses = -log_probs * advantages
    losses = tf.debugging.check_numerics(losses, "loss")
    return losses


def clipped_policy_gradient(
    log_probs, log_probs_anchor, advantages, epsilon_clipping=0.2
):
    """
    Computes the clipped surrogate objective found in
    [Proximal Policy Optimization](https://arxiv.org/abs/1707.06347) based on
    clipped probability ratios.

    Args:
        log_probs: Log probabilities of taking actions under a policy.
        log_probs_anchor: Log probabilities of taking actions under an anchor
            policy which is updated less frequently.
        advantages: Advantage estimation.
        epsilon_clipping: Scalar for clipping the policy ratio.
        sample_weight: Optional tensor for weighting the losses.

    Returns:
        Tensor of losses.
    """
    log_probs_anchor = tf.stop_gradient(log_probs_anchor)
    advantages = tf.stop_gradient(advantages)

    ratio = tf.exp(log_probs - log_probs_anchor)
    ratio_clipped = tf.clip_by_value(ratio, 1 - epsilon_clipping, 1 + epsilon_clipping)

    surrogate1 = ratio * advantages
    surrogate2 = ratio_clipped * advantages

    surrogate_min = tf.minimum(surrogate1, surrogate2)

    losses = -surrogate_min
    losses = tf.debugging.check_numerics(losses, "losses")
    return losses


def policy_entropy(entropy):
    """
    Computes the policy entropy loss.

    Args:
        entropy: Entropy of the policy distribution.

    Returns:
        Tensor of losses.
    """
    losses = -entropy
    losses = tf.debugging.check_numerics(losses, "losses")
    return losses


class PolicyGradient(tf.keras.losses.Loss):
    """
    Computes the Vanilla policy gradient loss.

    Attributes:
        reduction: a tf.keras.losses.Reduction method.
        name: name of the loss.
    """

    def __init__(
        self, reduction=tf.keras.losses.Reduction.SUM_OVER_BATCH_SIZE, name=None
    ):
        self.reduction = reduction
        self.name = name

    def __call__(self, log_probs, advantages, sample_weight=None):
        losses = policy_gradient(log_probs, advantages)
        return losses_utils.compute_weighted_loss(
            losses, sample_weight, reduction=self.reduction
        )

    @classmethod
    def from_config(cls, config):
        return cls(**config)

    def get_config(self):
        return {"reduction": self.reduction, "name": self.name}


class ClippedPolicyGradient(tf.keras.losses.Loss):
    """
    Computes the clipped surrogate objective found in
    [Proximal Policy Optimization](https://arxiv.org/abs/1707.06347) based on
    clipped probability ratios.

    Attributes:
        epsilon_clipping: epsilon parameter of the clipped surrogate objective.
        reduction: a tf.keras.losses.Reduction method.
        name: name of the loss.
    """

    def __init__(
        self,
        epsilon_clipping=0.2,
        reduction=tf.keras.losses.Reduction.SUM_OVER_BATCH_SIZE,
        name=None,
    ):
        self.epsilon_clipping = epsilon_clipping
        self.reduction = reduction
        self.name = name

    def __call__(self, log_probs, log_probs_anchor, advantages, sample_weight=None):
        losses = clipped_policy_gradient(log_probs, log_probs_anchor, advantages)
        return losses_utils.compute_weighted_loss(
            losses, sample_weight, reduction=self.reduction
        )

    @classmethod
    def from_config(cls, config):
        return cls(**config)

    def get_config(self):
        return {
            "epsilon_clipping": self.epsilon_clipping,
            "reduction": self.reduction,
            "name": self.name,
        }


class PolicyEntropy(tf.keras.losses.Loss):
    """
    Computes the policy entropy loss.

    Attributes:
        reduction: a tf.keras.losses.Reduction method.
        name: name of the loss.
    """

    def __init__(
        self, reduction=tf.keras.losses.Reduction.SUM_OVER_BATCH_SIZE, name=None
    ):
        self.reduction = reduction
        self.name = name

    def __call__(self, entropy, sample_weight=None):
        losses = policy_entropy(entropy)
        return losses_utils.compute_weighted_loss(
            losses, sample_weight, reduction=self.reduction
        )

    @classmethod
    def from_config(cls, config):
        return cls(**config)

    def get_config(self):
        return {"reduction": self.reduction, "name": self.name}



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from pyoneer.math import math_ops
from pyoneer.manip import array_ops, indexing_ops


class DiscountedReturns(object):
    """
    Compute discounted returns.

    Args:
        rewards: Rewards tensor.
        discount_factor: Weighting factor for discounting.
        sample_weight: Optional sample_weight tensor.

    Returns:
        Tensor of discounted returns.
    """

    def __init__(self, discount_factor=0.99):
        self.discount_factor = discount_factor

    def __call__(self, rewards, sample_weight=1.0):
        returns = tf.reverse(
            tf.transpose(
                tf.scan(
                    lambda agg, cur: cur + self.discount_factor * agg,
                    tf.transpose(tf.reverse(rewards * sample_weight, [1]), [1, 0]),
                    tf.zeros_like(rewards[:, -1]),
                    1,
                    False,
                ),
                [1, 0],
            ),
            [1],
        )
        returns = returns * sample_weight
        returns = tf.debugging.check_numerics(returns, "returns")
        returns = tf.stop_gradient(returns)
        return returns


class GeneralizedAdvantages(object):
    def __init__(self, discount_factor=0.99, lambda_factor=0.95, normalize=True):
        self.discount_factor = discount_factor
        self.lambda_factor = lambda_factor
        self.normalize = normalize

    def __call__(self, rewards, values, sample_weight=1.0):
        """
        Compute generalized advantage for policy optimization. Equation 11 and 12.

        Args:
            rewards: Rewards tensor.
            discount_factor: Weighting factor for discounting.
            sample_weight: Optional sample_weight tensor.

        Returns:
            Tensor of discounted rewards.
        """
        rewards = tf.convert_to_tensor(rewards)
        values = tf.convert_to_tensor(values)
        sample_weight = tf.convert_to_tensor(sample_weight)

        sequence_lengths = tf.reduce_sum(tf.ones_like(rewards) * sample_weight, axis=1)
        last_steps = tf.cast(sequence_lengths - 1, tf.int32)
        bootstrap_values = indexing_ops.batched_index(values, last_steps)

        values_next = array_ops.shift(
            values, shift=-1, axis=1, padding_values=bootstrap_values[:, None]
        )

        deltas = rewards + self.discount_factor * values_next - values

        advantages = tf.reverse(
            tf.transpose(
                tf.scan(
                    lambda agg, cur: cur
                    + self.discount_factor * self.lambda_factor * agg,
                    tf.transpose(tf.reverse(deltas * sample_weight, [1]), [1, 0]),
                    tf.zeros_like(deltas[:, -1]),
                    1,
                    False,
                ),
                [1, 0],
            ),
            [1],
        )

        if self.normalize:
            advantages_mean, advantages_variance = tf.nn.weighted_moments(
                advantages, axes=[0, 1], frequency_weights=sample_weight, keepdims=True
            )
            advantages_std = tf.sqrt(advantages_variance)
            advantages = math_ops.normalize(
                advantages,
                loc=advantages_mean,
                scale=advantages_std,
                sample_weight=sample_weight,
            )

        advantages = advantages * sample_weight
        advantages = tf.debugging.check_numerics(advantages, "advantages")
        advantages = tf.stop_gradient(advantages)
        return advantages



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import tensorflow as tf

from pyoneer.math import angle_ops
from pyoneer.layers.layers_impl import Swish, OneHotEncoder, AngleEncoder, Nest


class LayersTest(tf.test.TestCase):
    def test_swish_layer(self):
        layer = Swish()
        inputs = tf.constant([-1.0, 0.0, +1.0], dtype=tf.float32)
        outputs = layer(inputs)
        expected = tf.constant([-0.268941, 0.0, 0.731059], dtype=tf.float32)
        self.assertAllClose(outputs, expected)

    def test_one_hot_encoder(self):
        layer = OneHotEncoder(depth=4)
        inputs = tf.constant([3], dtype=tf.int32)
        outputs = layer(inputs)
        expected = tf.constant([[0.0, 0.0, 0.0, 1.0]], dtype=tf.float32)
        self.assertAllEqual(outputs, expected)

    def test_angle_encoder_radians(self):
        layer = AngleEncoder(degrees=False)
        inputs = tf.constant([[math.pi]], dtype=tf.float32)
        outputs = layer(inputs)
        expected = tf.constant([[-1, 0]], dtype=tf.float32)
        self.assertAllClose(outputs, expected)

    def test_angle_encoder_degrees(self):
        layer = AngleEncoder(degrees=True)
        inputs = tf.constant([[180]], dtype=tf.float32)
        outputs = layer(inputs)
        expected = tf.constant([[-1, 0]], dtype=tf.float32)
        self.assertAllClose(outputs, expected)

    def test_nest_features(self):
        nest_layer = Nest(
            {
                "category": OneHotEncoder(depth=3),
                "angle": AngleEncoder(degrees=True),
                "nested": {
                    "category": OneHotEncoder(depth=3),
                    "angle": AngleEncoder(degrees=True),
                },
            }
        )
        features = {
            "category": tf.constant([1]),
            "angle": tf.constant([[-45.0]]),
            "nested": {"category": tf.constant([2]), "angle": tf.constant([[45.0]])},
        }
        outputs = nest_layer(features)
        expected = tf.stack(
            [
                [
                    tf.math.cos(angle_ops.to_radians(-45.0)),
                    tf.math.sin(angle_ops.to_radians(-45.0)),
                    0.0,
                    1.0,
                    0.0,
                    tf.math.cos(angle_ops.to_radians(+45.0)),
                    tf.math.sin(angle_ops.to_radians(+45.0)),
                    0.0,
                    0.0,
                    1.0,
                ]
            ],
            axis=0,
        )
        self.assertAllClose(outputs, expected)


if __name__ == "__main__":
    tf.test.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from pyoneer.distributions.distributions_impl import MultiCategorical

__all__ = ["MultiCategorical"]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from pyoneer import activations
from pyoneer import debugging
from pyoneer import distributions
from pyoneer import initializers
from pyoneer import layers
from pyoneer import manip
from pyoneer import math
from pyoneer import metrics
from pyoneer import moments
from pyoneer import rl
from pyoneer import schedules

__all__ = [
    "activations",
    "debugging",
    "distributions",
    "initializers",
    "layers",
    "manip",
    "math",
    "metrics",
    "moments",
    "rl",
    "schedules",
]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf
import tensorflow_probability as tfp


class EpsilonGreedy(object):
    """
    Epsilon-greedy strategy. Samples from a policy distribution with
    `1 - epsilon` probability.

    Example:

        Decay ε every 1000 steps with a base of 0.96:

        ```
        global_step = tfe.Variable(0, trainable=False)
        initial_epsilon = 1.0 # only random samples
        strategy = EpsilonGreedyStrategy(
            tf.train.exponential_decay(
                initial_epsilon,
                global_step,
                1000,
                0.96))
        ```

    Args:
        policy: callable that returns a `tfp.distributions.Distribution`.
        epsilon: epsilon value. This can be
            a callable that takes no arguments and returns the actual
            value to use.
    """

    def __init__(self, policy, epsilon=1.0):
        self.policy = policy
        self.epsilon = epsilon

    def __call__(self, *args, **kwargs):
        policy = self.policy(*args, **kwargs)
        epsilon = self.epsilon() if callable(self.epsilon) else self.epsilon
        mask_dist = tfp.distributions.Bernoulli(probs=1 - epsilon, dtype=tf.bool)
        sample_mask = mask_dist.sample(policy.batch_shape)
        sample = policy.sample()
        mode = policy.mode()
        return tf.where(sample_mask, mode, sample)


class Mode(object):
    """
    Returns the mode of the policy distribution.

    Args:
        policy: callable that returns a `tfp.distributions.Distribution`.
    """

    def __init__(self, policy):
        self.policy = policy

    def __call__(self, *args, **kwargs):
        dist = self.policy(*args, **kwargs)
        return dist.mode()


class Sample(object):
    """
    Returns random samples from the policy distribution.

    Args:
        policy: callable that returns a `tfp.distributions.Distribution`.
    """

    def __init__(self, policy):
        self.policy = policy

    def __call__(self, *args, **kwargs):
        dist = self.policy(*args, **kwargs)
        return dist.sample()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf
import tensorflow_probability as tfp

from pyoneer.distributions.distributions_impl import MultiCategorical


class DistributionsTest(tf.test.TestCase):
    def test_multicategorical_log_prob(self):
        dist1 = tfp.distributions.Categorical(logits=[[0.1, 0.2, 0.3]])
        dist2 = tfp.distributions.Categorical(logits=[[0.3, 0.2, 0.1]])

        distribution = MultiCategorical([dist1, dist2])
        output = distribution.log_prob([[2, 0]])

        expected = tf.constant([-2.0038857], dtype=tf.float32)
        self.assertAllClose(output, expected)

    def test_multicategorical_entropy(self):
        dist1 = tfp.distributions.Categorical(logits=[[0.1, 0.2, 0.3]])
        dist2 = tfp.distributions.Categorical(logits=[[0.3, 0.2, 0.1]])

        distribution = MultiCategorical([dist1, dist2])
        output = distribution.entropy()

        expected = tf.constant([2.1905746], dtype=tf.float32)
        self.assertAllClose(output, expected)

    def test_multicategorical_mode(self):
        dist1 = tfp.distributions.Categorical(logits=[[0.1, 0.2, 0.3]])
        dist2 = tfp.distributions.Categorical(logits=[[0.3, 0.2, 0.1]])

        distribution = MultiCategorical([dist1, dist2])
        output = distribution.mode()

        expected = tf.constant([[2, 0]], dtype=tf.int32)
        self.assertAllClose(output, expected)


if __name__ == "__main__":
    tf.test.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf


class Stopwatch(object):
    """
    Stopwatch for measuring how long operations take. Great for fast and easy profiling.

    Example:
    >>> x = tf.constant(1.0)
    >>> y = tf.constant(2.0)
    >>> with Stopwatch() as watch:
    >>>    z = x + y
    >>> tf.print(watch.duration)
    >>> # 0.00021505355834960938
    """

    def __init__(self):
        self.start_time = None
        self.end_time = None
        self.duration = None

    def start(self):
        self.start_time = tf.timestamp()

    def stop(self):
        self.end_time = tf.timestamp()
        self.duration = self.end_time - self.start_time

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from pyoneer.debugging.debugging_impl import Stopwatch


class DebuggingTest(tf.test.TestCase):
    def test_stopwatch(self):
        with Stopwatch() as stopwatch:
            pass
        self.assertIsNotNone(stopwatch.start_time)
        self.assertIsNotNone(stopwatch.end_time)
        self.assertIsNotNone(stopwatch.duration)


if __name__ == "__main__":
    tf.test.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from setuptools import setup, find_packages

install_requires = ["gym", "six"]
with open("README.md", "r") as fp:
    long_description = fp.read()

setup(
    name="fomoro-pyoneer",
    version="0.3.0",
    author="Fomoro AI",
    author_email="team@fomoro.com",
    description="Tensor utilities, reinforcement learning, and more!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fomorians/pyoneer",
    download_url="https://github.com/fomorians/pyoneer/archive/v0.3.tar.gz",
    license="Apache 2.0",
    packages=find_packages(),
    install_requires=install_requires,
    extras_require={"dev": ["pre-commit"]},
    include_package_data=True,
    keywords=[
        "tensorflow",
        "machine learning",
        "reinforcement learning",
        "eager execution",
        "deep learning",
    ],
    classifiers=[
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
    ],
)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from pyoneer.initializers.initializers_impl import SoftplusInverse

__all__ = ["SoftplusInverse"]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from pyoneer.rl.strategies.strategies_impl import EpsilonGreedy, Mode, Sample

__all__ = ["EpsilonGreedy", "Mode", "Sample"]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf


def swish(x):
    """
    Compute Swish, self-gating, activation function: `x * sigmoid(x)`.

    Args:
        x: Tensor

    Returns:
        Tensor of same dimension as `x`.
    """
    y = x * tf.sigmoid(x)
    y = tf.debugging.check_numerics(y, "swish")
    return y



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf


def flatten(tensor):
    """
    Flatten a tensor by reshaping into a vector.

    Args:
        tensor: Tensor to flatten.

    Returns:
        Flattened tensor.
    """
    return tf.reshape(tensor, [-1])


def pad_or_truncate(tensor, sizes, mode="CONSTANT", constant_values=0.0):
    """
    Pad or truncate a tensor. This is useful for ensuring sequences have the
    same minimum shape when the sequence length is not known in advance.

    Examples:

    sentences = [['apple', 'is', 'red'], ['i', 'dont', 'like', 'apples']]
    sentences = tf.stack([
        pynr.manip.pad_or_truncate(sentences, sizes=[2, 3], '<PAD>')
        for s in sentences], axis=0)
    >>> [['apple', 'is', 'red'], ['i', 'dont', 'like']]

    sentences = [['apple', 'is', 'red'], ['i', 'dont', 'like', 'apples']]
    sentences = tf.stack([
        pynr.manip.pad_or_truncate(sentences, sizes=[2, 4], '<PAD>')
        for s in sentences], axis=0)
    >>> [['apple', 'is', 'red', '<PAD>'], ['i', 'dont', 'like', 'apples']]

    Args:
        tensor: Tensor to pad or truncate.
        sizes: Maximum sizes for tensor along each axis. If tensor.shape is
            greater than sizes along any axis, the axis will be trucated,
            otherwise they will be padded according to `mode` and
            `constant_values`.
        mode: One of "CONSTANT", "REFLECT", or "SYMMETRIC" (case-insensitive).
        constant_values: In "CONSTANT" mode, the padding value to use.

    Returns:
        Padded or truncated tensor.
    """
    tensor = tf.convert_to_tensor(tensor)

    paddings = [
        [0, max(sizes[axis] - tensor.shape[axis], 0)]
        for axis in range(tensor.shape.ndims)
    ]

    slices = [slice(0, sizes[axis]) for axis in range(tensor.shape.ndims)]

    tensor = tf.pad(
        tensor=tensor, paddings=paddings, mode=mode, constant_values=constant_values
    )
    tensor = tensor[slices]

    return tensor


def shift(inputs, shift, axis, padding_values=0.0):
    """
    Shifts the elements of a tensor along an axis.

    This is similar to `tf.manip.roll`, except that it optionally fills the
    rolled Tensors with `padding_values`.

    Examples:

        Convert states to next states and fill last step with zeros:

            next_state = pynr.manip.shift(states, shift=-1, axis=1)

        Convert values to next values and use bootstrap values for last step:

            next_values = pynr.manip.shift(values, shift=-1, axis=1,
                padding_values=bootstrap_values[:, None])

        Convert actions and rewards to previous actions and rewards for RL^2:

            previous_actions = pynr.manip.shift(actions, shift=1, axis=1)
            previous_rewards = pynr.manip.shift(rewards, shift=1, axis=1)

    Args:
        inputs: A `Tensor` to shift.
        shift:  A `Tensor`. Must be one of the following types: int32, int64.
            Dimension must be 0-D or 1-D. shift[i] specifies the number of
            places by which elements are shifted positively (towards larger
            indices) along the dimension specified by axis[i]. Negative shifts
            will roll the elements in the opposite direction.
        axis: A `Tensor`. Must be one of the following types: int32, int64.
            Dimension must be 0-D or 1-D. axis[i] specifies the dimension that
            the shift shift[i] should occur. If the same axis is referenced
            more than once, the total shift for that axis will be the sum of
            all the shifts that belong to that axis.
        padding_values: A tensor to use for padding.

    Returns:
        A shifted Tensor.
    """
    inputs = tf.convert_to_tensor(inputs)

    # slice inputs
    slices = [slice(None)] * inputs.shape.ndims

    if shift >= 0:
        slices[axis] = slice(0, -shift)
    else:
        slices[axis] = slice(-shift, None)

    sliced = inputs[slices]

    # pad inputs
    paddings = [[0, 0]] * inputs.shape.ndims
    paddings[axis] = [max(shift, 0), max(-shift, 0)]
    padding_slices = [slice(None)] * inputs.shape.ndims

    if shift >= 0:
        padding_slices[axis] = slice(-shift, None)
        padded = tf.concat(
            [padding_values * tf.ones_like(inputs[padding_slices]), sliced], axis=axis
        )
    else:
        padding_slices[axis] = slice(0, -shift)
        padded = tf.concat(
            [sliced, tf.ones_like(inputs[padding_slices]) * padding_values], axis=axis
        )

    return padded



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf


class MultiCategorical(object):
    """
    Distribution composed of multiple distributions.

    Useful for representing `gym.spaces.MultiDiscrete`.

    Args:
        distributions: list of distributions.
    """

    def __init__(self, distributions):
        self.distributions = distributions

    def log_prob(self, value):
        values = tf.split(value, len(self.distributions), axis=-1)
        log_probs = [
            dist.log_prob(val[..., 0]) for dist, val in zip(self.distributions, values)
        ]
        return tf.math.add_n(log_probs)

    def entropy(self):
        return tf.math.add_n([dist.entropy() for dist in self.distributions])

    def sample(self):
        return tf.stack([dist.sample() for dist in self.distributions], axis=-1)

    def mode(self):
        return tf.stack([dist.mode() for dist in self.distributions], axis=-1)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from pyoneer.layers.layers_impl import Swish, OneHotEncoder, AngleEncoder, Nest

__all__ = ["Swish", "OneHotEncoder", "AngleEncoder", "Nest"]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from pyoneer.schedules.cyclic_schedule_impl import CyclicSchedule


class CyclicScheduleTest(tf.test.TestCase):
    def test_cyclic_schedule(self):
        cyclic_lr = CyclicSchedule(
            minval=0.2, maxval=0.4, step_size=500, decay_to_zero=False
        )
        # start of a cycle
        self.assertAllClose(cyclic_lr(0), 0.2)
        # apex of the cycle
        self.assertAllClose(cyclic_lr(500), 0.4)
        # midway from the apex to the end of a cycle
        self.assertAllClose(cyclic_lr(750), 0.3)
        # end of a cycle
        self.assertAllClose(cyclic_lr(999), 0.2004)
        # start of a new cycle
        self.assertAllClose(cyclic_lr(1000), 0.2)

    def test_cyclic_schedule_decay(self):
        cyclic_lr = CyclicSchedule(
            minval=0.2, maxval=0.4, step_size=500, decay_to_zero=True
        )
        # start of a cycle
        self.assertAllClose(cyclic_lr(0), 0.2)
        # apex of the cycle
        self.assertAllClose(cyclic_lr(500), 0.4)
        # midway from the apex to the end of a cycle
        self.assertAllClose(cyclic_lr(750), 0.2)
        # end of a cycle
        self.assertAllClose(cyclic_lr(999), 0.0008)
        # start of a new cycle
        self.assertAllClose(cyclic_lr(1000), 0.2)


if __name__ == "__main__":
    tf.test.main()
